import numpy as np
import pandas as pd
from odeintw import odeintw
from functools import partial
import matplotlib.pylab as plt
import time
from numba import jit, njit
from cmath import exp, sqrt, pi



plt.style.use("mundus")


def R(u):
    """
    Auxiliary R function that appears in the windowing function c_inf_window.
    """
    return np.piecewise(u,[u<=0, u>0],[0,  lambda u : np.exp(-1/u)])


def S(u):
    """
    Auxiliary S function that appears in the windowing function c_inf_window.
    """
    return R(u)/(R(u)+R(1-u))


def cinf_window_old(t0, T, tau, A, t):
    u = (t - t0)/tau
    y = A*S(u)*S(T/tau-u)
    return y


def zero_window(t):
    return 0


@njit
def cinf_window(t0, T, tau, A, t):
    u = (t - t0) / tau

    p = u
    if p > 0:
        Rp = np.exp(-1 / p)
    else:
        Rp = 0

    one_minus_p = 1 - u
    if one_minus_p > 0:
        Rone_minus_p = np.exp(-1 / one_minus_p)
    else:
        Rone_minus_p = 0

    T_per_tau_minus_p = T / tau - u
    if T_per_tau_minus_p > 0:
        RT_per_tau_minus_p = np.exp(-1 / T_per_tau_minus_p)
    else:
        RT_per_tau_minus_p = 0

    one_minusT_per_tau_minus_p = 1 - (T / tau - u)
    if one_minusT_per_tau_minus_p > 0:
        Rone_minusT_per_tau_minus_p = np.exp(-1 / one_minusT_per_tau_minus_p)
    else:
        Rone_minusT_per_tau_minus_p = 0

    Sp = Rp / (Rp + Rone_minus_p)
    ST_per_tau_minus_p = RT_per_tau_minus_p / (RT_per_tau_minus_p + Rone_minusT_per_tau_minus_p)
    return A * Sp * ST_per_tau_minus_p


#@profile
def rate_equations5modes(b_,t,
                         omega0, omega1, omega2, omega3, omega4,
                         Gamma0, Gamma1, Gamma2, Gamma3, Gamma4,
                         W0, W1, W2, W3, W4,
                         V, Tkk,
                         therm0, therm1, therm2, therm3, therm4):
    """
    Rate equations for the canonical variables b0 to b4 including
    pumping as well as three-magnon and four-magnon scattering. This function
    takes all necessary parameters and returns the time derivatives of the
    canonical variables.

    parameters:
        b_      : complex valued array of length 5 containing the mode variables
        t       : time in ns
        omega_  : real valued array of length 5 containing the mode frequencies
        Gamma_  : real valued array of length 5 containing the mode life times
        W_       : Windowing functions for the field, list of functions
        V       : Three-magnon scattering coefficient
        Tkk     : Four-magnon scattering coefficient

    returns:
        dbdt_ : list of length 5 containg time derivatives of the mode variables
    """

    # match initial values and parameters.
    b0, b1, b2, b3, b4 = b_

    # Time derivatives

    # direct mode
    W0_t = W0(t)
    db0dt = -(Gamma0 + 1j*omega0)*b0 + Gamma0*sqrt(therm0) * \
        exp(-1j*omega0*t)*1j + \
            1j*W0_t*exp(-1j*omega0*t) - \
            1j*V*b1*b2 - 1j*V*b3*b4

    #np.exp(-1j*omega0*t)
    # n1, +m
    W1_t = W1(t)
    db1dt = -(Gamma1 + 1j*(omega1 + 2*Tkk*(abs(b1)**2)))*b1 + \
        Gamma1*sqrt(therm1)*exp(-1j*omega1*t)*1j - \
        1j*V.conjugate()*b2.conjugate()*b0 + \
            1j*W1_t*exp(-1j*omega1*t)

    # n2, -m
    W2_t = W2(t)
    db2dt = -(Gamma2 + 1j*(omega2 + 2*Tkk*(abs(b2)**2)))*b2 + \
        Gamma2*sqrt(therm2)*exp(-1j*omega2*t)*1j - \
        1j*V.conjugate()*b1.conjugate()*b0 + \
            1j*W2_t*exp(-1j*omega2*t)

    # n1, -m
    W3_t = W3(t)
    db3dt = -(Gamma3 + 1j*(omega3 + 2*Tkk*(abs(b3)**2)))*b3 + \
        Gamma3*sqrt(therm3)*exp(-1j*omega3*t)*1j - \
        1j*V.conjugate()*b4.conjugate()*b0 + \
            1j*W3_t*exp(-1j*omega3*t)

    # n2, +m
    W4_t = W4(t)
    db4dt = -(Gamma4 + 1j*(omega4 + 2*Tkk*(abs(b4)**2)))*b4 + \
        Gamma4*sqrt(therm4)*exp(-1j*omega4*t)*1j - \
        1j*V.conjugate()*b3.conjugate()*b0 + \
            1j*W4_t*exp(-1j*omega4*t)

    return [db0dt, db1dt, db2dt, db3dt, db4dt]


def rate_equations5modes_differentpairs(b_,t,
                         omega0, omega1, omega2, omega3, omega4,
                         Gamma0, Gamma1, Gamma2, Gamma3, Gamma4,
                         W0, W1, W2, W3, W4,
                         V1, V2, Tkk,
                         therm0, therm1, therm2, therm3, therm4):
    """
    Rate equations for the canonical variables b0 to b4 including
    pumping as well as three-magnon and four-magnon scattering. This function
    takes all necessary parameters and returns the time derivatives of the
    canonical variables.

    parameters:
        b_      : complex valued array of length 5 containing the mode variables
        t       : time in ns
        omega_  : real valued array of length 5 containing the mode frequencies
        Gamma_  : real valued array of length 5 containing the mode life times
        W_       : Windowing functions for the field, list of functions
        V       : Three-magnon scattering coefficient
        Tkk     : Four-magnon scattering coefficient

    returns:
        dbdt_ : list of length 5 containg time derivatives of the mode variables
    """

    # match initial values and parameters.
    b0, b1, b2, b3, b4 = b_

    # Time derivatives

    # direct mode
    W0_t = W0(t)
    db0dt = -(Gamma0 + 1j*omega0)*b0 + Gamma0*sqrt(therm0) * \
        exp(-1j*omega0*t)*1j + \
            1j*W0_t*exp(-1j*omega0*t) - \
            1j*V1*b1*b2 - 1j*V2*b3*b4

    #np.exp(-1j*omega0*t)
    # n1, +m
    W1_t = W1(t)
    db1dt = -(Gamma1 + 1j*(omega1 + 2*Tkk*(abs(b1)**2)))*b1 + \
        Gamma1*sqrt(therm1)*exp(-1j*omega1*t)*1j - \
        1j*V1.conjugate()*b2.conjugate()*b0 + \
            1j*W1_t*exp(-1j*omega1*t)

    # n2, -m
    W2_t = W2(t)
    db2dt = -(Gamma2 + 1j*(omega2 + 2*Tkk*(abs(b2)**2)))*b2 + \
        Gamma2*sqrt(therm2)*exp(-1j*omega2*t)*1j - \
        1j*V1.conjugate()*b1.conjugate()*b0 + \
            1j*W2_t*exp(-1j*omega2*t)

    # n1, -m
    W3_t = W3(t)
    db3dt = -(Gamma3 + 1j*(omega3 + 2*Tkk*(abs(b3)**2)))*b3 + \
        Gamma3*sqrt(therm3)*exp(-1j*omega3*t)*1j - \
        1j*V2.conjugate()*b4.conjugate()*b0 + \
            1j*W3_t*exp(-1j*omega3*t)

    # n2, +m
    W4_t = W4(t)
    db4dt = -(Gamma4 + 1j*(omega4 + 2*Tkk*(abs(b4)**2)))*b4 + \
        Gamma4*sqrt(therm4)*exp(-1j*omega4*t)*1j - \
        1j*V2.conjugate()*b3.conjugate()*b0 + \
            1j*W4_t*exp(-1j*omega4*t)

    return [db0dt, db1dt, db2dt, db3dt, db4dt]



def rate_equations5modes_Sterms(b_,t,
                         omega0, omega1, omega2, omega3, omega4,
                         Gamma0, Gamma1, Gamma2, Gamma3, Gamma4,
                         W0, W1, W2, W3, W4,
                         V, Tkk, Skk,
                         therm0, therm1, therm2, therm3, therm4):
    """
    Rate equations for the canonical variables b0 to b4 including
    pumping as well as three-magnon and four-magnon scattering. This function
    takes all necessary parameters and returns the time derivatives of the
    canonical variables.

    parameters:
        b_      : complex valued array of length 5 containing the mode variables
        t       : time in ns
        omega_  : real valued array of length 5 containing the mode frequencies
        Gamma_  : real valued array of length 5 containing the mode life times
        W_       : Windowing functions for the field, list of functions
        V       : Three-magnon scattering coefficient
        Tkk     : Four-magnon scattering coefficient

    returns:
        dbdt_ : list of length 5 containg time derivatives of the mode variables
    """

    global omegaNL1
    global t_omega
    global omegaNL2

    # match initial values and parameters.
    b0, b1, b2, b3, b4 = b_

    # Time derivatives

    # direct mode
    W0_t = W0(t)
    db0dt = -(Gamma0 + 1j*omega0)*b0 + Gamma0*sqrt(therm0) * \
        exp(-1j*omega0*t)*1j + \
            1j*W0_t*exp(-1j*omega0*t) - \
            1j*V*b1*b2 - 1j*V*b3*b4

    #np.exp(-1j*omega0*t)
    # n1, +m
    W1_t = W1(t)
    omegaNL_1 = (omega1 - 2*Tkk*(abs(b1)**2))
    omegaNL_2 = (omega2 + 2*Tkk*(abs(b2)**2))
    omegaNL1.append(omegaNL_1/(2*np.pi))
    t_omega.append(t)
    omegaNL2.append(omegaNL_2 / (2 * np.pi))


    db1dt = -(Gamma1 + 1j*(omegaNL_1))*b1 + \
        Gamma1*sqrt(therm1)*exp(-1j*omega1*t)*1j - \
        1j*V.conjugate()*b2.conjugate()*b0 + \
            1j*W1_t*exp(-1j*omega1*t) + \
            -2j*Tkk*(abs(b3)**2)*b1 + \
            -1j*Skk*(b3**2)*b1.conjugate()

    # n2, -m
    W2_t = W2(t)
    db2dt = -(Gamma2 + 1j*(omega2 + 2*Tkk*(abs(b2)**2)))*b2 + \
        Gamma2*sqrt(therm2)*exp(-1j*omega2*t)*1j - \
        1j*V.conjugate()*b1.conjugate()*b0 + \
            1j*W2_t*exp(-1j*omega2*t) + \
            -2j*Tkk*(abs(b4)**2)*b2 + \
            -1j*Skk*(b4**2)*b2.conjugate()

    # n1, -m
    W3_t = W3(t)
    db3dt = -(Gamma3 + 1j*(omega3 - 2*Tkk*(abs(b3)**2)))*b3 + \
        Gamma3*sqrt(therm3)*exp(-1j*omega3*t)*1j - \
        1j*V.conjugate()*b4.conjugate()*b0 + \
            1j*W3_t*exp(-1j*omega3*t) + \
            -2j*Tkk*(abs(b1)**2)*b3 + \
            -1j*Skk*(b1**2)*b3.conjugate()

    # n2, +m
    W4_t = W4(t)
    db4dt = -(Gamma4 + 1j*(omega4 + 2*Tkk*(abs(b4)**2)))*b4 + \
        Gamma4*sqrt(therm4)*exp(-1j*omega4*t)*1j - \
        1j*V.conjugate()*b3.conjugate()*b0 + \
            1j*W4_t*exp(-1j*omega4*t) + \
            -2j*Tkk*(abs(b2)**2)*b4 + \
            -1j*Skk*(b2**2)*b4.conjugate()

    return [db0dt, db1dt, db2dt, db3dt, db4dt]


def rate_equations5modes_nonlineardamping(b_,t,
                         omega0, omega1, omega2, omega3, omega4,
                         Gamma0, Gamma1, Gamma2, Gamma3, Gamma4,
                         W0, W1, W2, W3, W4,
                         V, Tkk, eta,
                         therm0, therm1, therm2, therm3, therm4):
    """
    Rate equations for the canonical variables b0 to b4 including
    pumping as well as three-magnon and four-magnon scattering. This function
    takes all necessary parameters and returns the time derivatives of the
    canonical variables.

    parameters:
        b_      : complex valued array of length 5 containing the mode variables
        t       : time in ns
        omega_  : real valued array of length 5 containing the mode frequencies
        Gamma_  : real valued array of length 5 containing the mode life times
        W_       : Windowing functions for the field, list of functions
        V       : Three-magnon scattering coefficient
        Tkk     : Four-magnon scattering coefficient

    returns:
        dbdt_ : list of length 5 containg time derivatives of the mode variables
    """

    # match initial values and parameters.
    b0, b1, b2, b3, b4 = b_

    # Time derivatives

    # direct mode
    W0_t = W0(t)
    db0dt = -(Gamma0 + 1j*omega0)*b0 + Gamma0*sqrt(therm0) * \
        exp(-1j*omega0*t)*1j + \
            1j*W0_t*exp(-1j*omega0*t) - \
            1j*V*b1*b2 - 1j*V*b3*b4 - \
            eta * (abs(b0)**2 + abs(b1)**2 + abs(b2)**2 + abs(b3)**2 + abs(b4)**2)*b0

    #np.exp(-1j*omega0*t)
    # n1, +m
    W1_t = W1(t)
    db1dt = -(Gamma1 + 1j*(omega1 + 2*Tkk*(abs(b1)**2)))*b1 + \
        Gamma1*sqrt(therm1)*exp(-1j*omega1*t)*1j - \
        1j*V.conjugate()*b2.conjugate()*b0 + \
            1j*W1_t*exp(-1j*omega1*t) - \
            eta * (abs(b0)**2 + abs(b1)**2 + abs(b2)**2 + abs(b3)**2 + abs(b4)**2)*b1

    # n2, -m
    W2_t = W2(t)
    db2dt = -(Gamma2 + 1j*(omega2 + 2*Tkk*(abs(b2)**2)))*b2 + \
        Gamma2*sqrt(therm2)*exp(-1j*omega2*t)*1j - \
        1j*V.conjugate()*b1.conjugate()*b0 + \
            1j*W2_t*exp(-1j*omega2*t) - \
            eta * (abs(b0)**2 + abs(b1)**2 + abs(b2)**2 + abs(b3)**2 + abs(b4)**2)*b2

    # n1, -m
    W3_t = W3(t)
    db3dt = -(Gamma3 + 1j*(omega3 + 2*Tkk*(abs(b3)**2)))*b3 + \
        Gamma3*sqrt(therm3)*exp(-1j*omega3*t)*1j - \
        1j*V.conjugate()*b4.conjugate()*b0 + \
            1j*W3_t*exp(-1j*omega3*t) - \
            eta * (abs(b0)**2 + abs(b1)**2 + abs(b2)**2 + abs(b3)**2 + abs(b4)**2)*b3

    # n2, +m
    W4_t = W4(t)
    db4dt = -(Gamma4 + 1j*(omega4 + 2*Tkk*(abs(b4)**2)))*b4 + \
        Gamma4*sqrt(therm4)*exp(-1j*omega4*t)*1j - \
        1j*V.conjugate()*b3.conjugate()*b0 + \
            1j*W4_t*exp(-1j*omega4*t) - \
            eta * (abs(b0)**2 + abs(b1)**2 + abs(b2)**2 + abs(b3)**2 + abs(b4)**2)*b4

    return [db0dt, db1dt, db2dt, db3dt, db4dt]

def rate_equations5modes_J(b_,t,
                         omega0, omega1, omega2, omega3, omega4,
                         Gamma0, Gamma1, Gamma2, Gamma3, Gamma4,
                         W0, W1, W2, W3, W4,
                         V, Tkk,
                         therm0, therm1, therm2, therm3, therm4):
    # match initial values and parameters.
    b0, b1, b2, b3, b4 = b_

    # Time derivatives

    # direct mode
    W0_t = W0(t)

    J0 = [-(Gamma0 + 1j*omega0),
          1j*V*b2,
          1j*V*b1,
          -1j*V*b4,
          -1j*V*b3
          ]

    J1 = [-1j*V.conjugate()*b2.conjugate(),
          -(Gamma1 + 1j*(omega1 + 4*Tkk*(abs(b1)**2))),
          0,
          0,
          0]

    J2 = [-1j*V.conjugate()*b1.conjugate(),
          0,
          -(Gamma2 + 1j * (omega2 + 4 * Tkk * (abs(b2) ** 2))),
          0,
          0,
          ]

    J3 = [-1j*V.conjugate()*b4.conjugate(),
          0,
          -(Gamma3 + 1j*(omega3 + 4*Tkk*(abs(b3)**2))),
          0,
          0]

    J4 = [-1j * V.conjugate() * b3.conjugate(),
          0,
          0,
          -(Gamma4 + 1j * (omega4 + 4 * Tkk * (abs(b4) ** 2))),
          0]


    return [J0, J1, J2, J3, J4]

def rate_equations5modes_numpy(b_,t, omega_, Gamma_, W_, V, Tkk, thermal_level_):
    """
    Rate equations for the canonical variables b0 to b4 including
    pumping as well as three-magnon and four-magnon scattering. This function
    takes all necessary parameters and returns the time derivatives of the
    canonical variables.

    parameters:
        b_      : complex valued array of length 5 containing the mode variables
        t       : time in ns
        omega_  : real valued array of length 5 containing the mode frequencies
        Gamma_  : real valued array of length 5 containing the mode life times
        W_       : Windowing functions for the field, list of functions
        V       : Three-magnon scattering coefficient
        Tkk     : Four-magnon scattering coefficient

    returns:
        dbdt_ : list of length 5 containg time derivatives of the mode variables
    """

    # print(time.clock(), t)

    # match initial values and parameters.
    b0, b1, b2, b3, b4 = b_
    omega0, omega1, omega2, omega3, omega4 = omega_
    Gamma0, Gamma1, Gamma2, Gamma3, Gamma4 = Gamma_
    W0, W1, W2, W3, W4 = W_
    therm0, therm1, therm2, therm3, therm4 = thermal_level_

    # Time derivatives

    # direct mode
    db0dt = -(Gamma0 + 1j*omega0)*b0 + Gamma0*np.sqrt(therm0) * \
        np.exp(-1j*omega0*t)*1j + \
            1j*W0(t)*np.exp(-1j*omega0*t) - \
            1j*V*b1*b2 - 1j*V*b3*b4

    #np.exp(-1j*omega0*t)
    # n1, +m
    db1dt = -(Gamma1 + 1j*(omega1 + 2*Tkk*(np.abs(b1)**2)))*b1 + \
        Gamma1*np.sqrt(therm1)*np.exp(-1j*omega1*t)*1j - \
        1j*np.conjugate(V)*np.conjugate(b2)*b0 + \
            1j*W1(t)*np.exp(-1j*omega1*t)

    # n2, -m
    db2dt = -(Gamma2 + 1j*(omega2 + 2*Tkk*(np.abs(b2)**2)))*b2 + \
        Gamma2*np.sqrt(therm2)*np.exp(-1j*omega2*t)*1j - \
        1j*np.conjugate(V)*np.conjugate(b1)*b0 + \
            1j*W2(t)*np.exp(-1j*omega2*t)

    # n1, -m
    W3_t = W3(t)
    db3dt = -(Gamma3 + 1j*(omega3 + 2*Tkk*(np.abs(b3)**2)))*b3 + \
        Gamma3*np.sqrt(therm3)*np.exp(-1j*omega3*t)*1j - \
        1j*np.conjugate(V)*np.conjugate(b4)*b0  + \
            1j*W3_t*np.exp(-1j*omega3*t)

    # n2, +m
    W4_t = W4(t)
    db4dt = -(Gamma4 + 1j*(omega4 + 2*Tkk*(np.abs(b4)**2)))*b4 + \
        Gamma4*np.sqrt(therm4)*np.exp(-1j*omega4*t)*1j - \
        1j*np.conjugate(V)*np.conjugate(b3)*b0 + \
            1j*W4_t*np.exp(-1j*omega4*t)

    return [db0dt, db1dt, db2dt, db3dt, db4dt]

omegaNL1 = []
t_omega = []
omegaNL2 = []